# 3-Clause BSD License
# 
# Copyright (c) 2009, Boxed Ice <hello@boxedice.com>
# Copyright (c) 2010-2016, Datadog <info@datadoghq.com>
# Copyright (c) 2020-present, Akita Software <info@akitasoftware.com>
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import copy
import time

from .har import *
from django.core.handlers.wsgi import WSGIRequest
from django.test import Client as DjangoClient
from urllib import parse


class Client(DjangoClient):
    """
    A wrapper around django.test.Client, which logs requests and responses to
    a HAR file.
    """

    def __init__(self, *args, har_file_path=f'akita_trace_{time.time()}.har', **kwargs):
        super().__init__(*args, **kwargs)
        self.har_writer = HarWriter(har_file_path, 'w')

    def request(self, **request):
        start = datetime.datetime.now(datetime.timezone.utc)
        response = super().request(**copy.deepcopy(request))
        self.har_writer.write_entry(self._entry_of_wsgi(start, request, response))

        return response

    def close(self):
        self.har_writer.close()

    def _entry_of_wsgi(self, start: datetime, request, response) -> Entry:
        wsgi = WSGIRequest(self._base_environ(**request))

        # Build request
        server_protocol = 'HTTP/1.1'
        if 'SERVER_PROTOCOL' in wsgi.environ:
            server_protocol = wsgi.environ['SERVER_PROTOCOL']

        url = parse.urlsplit(wsgi.get_raw_uri())
        query_string = [Record(name=k, value=v) for k, v in parse.parse_qs(url.query).items()]

        headers = [Record(name=k, value=v) for k, v in wsgi.headers.items()]
        encoded_headers = '\n'.join([f'{k}: {v}' for k, v in wsgi.headers.items()]).encode("utf-8")
        body = wsgi.body.decode("utf-8")

        har_request = Request(
            method=wsgi.method,
            url=url.path,
            httpVersion=server_protocol,
            cookies=[Record(name=k, value=v) for k, v in wsgi.COOKIES.items()],
            headers=headers,
            queryString=query_string,
            postData=None if not body else PostData(mimeType=wsgi.content_type, text=body),
            headersSize=len(encoded_headers),
            bodySize=len(wsgi.body),
        )

        # Build response
        content = response.content.decode("utf-8") if response.content is not None else ''
        headers = {}
        serialized_headers = response.serialize_headers()
        for x in serialized_headers.decode("utf-8").split('\r\n'):
            kv = x.split(':')
            headers[kv[0].strip()] = kv[1].strip()
        har_response = Response(
            status=response.status_code,
            statusText=response.reason_phrase,
            httpVersion=server_protocol,
            cookies=[Record(name=k, value=v.value) for k, v in response.cookies.items()],
            headers=[Record(name=k, value=v) for k, v in headers.items()],
            content=ResponseContent(size=len(content), mimeType=response.get('Content-Type', ''), text=content),
            redirectURL=response.url if 'url' in dir(response) else '',
            headersSize=len(serialized_headers),
            bodySize=len(response.content),
        )

        return Entry(
            startedDateTime=start,
            time=(datetime.datetime.now(datetime.timezone.utc) - start).total_seconds(),
            request=har_request,
            response=har_response,
            cache=Cache(),
            timings=Timings(send=0, wait=0, receive=0),
        )



# 3-Clause BSD License
# 
# Copyright (c) 2009, Boxed Ice <hello@boxedice.com>
# Copyright (c) 2010-2016, Datadog <info@datadoghq.com>
# Copyright (c) 2020-present, Akita Software <info@akitasoftware.com>
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import datetime
import json
import time

from . import __version__
from pydantic import BaseModel, validator
from queue import Empty, Queue
from threading import Thread
from typing import List, Optional, Union


class Number:
    """
    Type of a JSON 'number', which may be an int or float.
    """
    v: Union[int, float]

    def __init__(self, v: Union[int, float]):
        self.v = v

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if '.' in str(v):
            return float(v)
        return int(v)

    def __repr__(self):
        return str(self.v)


class Record(BaseModel):
    name: str
    value: str


class Timings(BaseModel):
    dns: Optional[Number]
    connect: Optional[Number]
    blocked: Optional[Number]
    send: Number
    wait: Number
    receive: Number
    ssl: Optional[Number]


class PostDataParam(BaseModel):
    name: str
    value: Optional[str]
    contentType: Optional[str]


class PostData(BaseModel):
    mimeType: str
    text: str
    params: Optional[List[PostDataParam]]


class ResponseContent(BaseModel):
    size: int
    mimeType: str
    text: str
    compression: Optional[int]
    encoding: Optional[str]


class Request(BaseModel):
    method: str
    url: str
    httpVersion: str
    cookies: List[Record]
    headers: List[Record]
    queryString: List[Record]
    postData: Optional[PostData]
    headersSize: int
    bodySize: int


class Response(BaseModel):
    status: int
    statusText: str
    httpVersion: str
    cookies: List[Record]
    headers: List[Record]
    content: ResponseContent
    redirectURL: str
    headersSize: int
    bodySize: int


class CacheEntry(BaseModel):
    expires: Optional[str]
    lastAccess: str
    eTag: str
    hitCount: int
    comment: Optional[str]


class Cache(BaseModel):
    beforeRequest: Optional[CacheEntry]
    afterRequest: Optional[CacheEntry]
    comment: Optional[str]


class Entry(BaseModel):
    startedDateTime: datetime.datetime
    time: Number
    request: Request
    response: Response
    cache: Cache
    timings: Timings

    @validator('startedDateTime')
    def datetime_must_have_timezone(cls, v: datetime.datetime):
        if v.tzinfo is None:
            raise ValueError('datetime must be timezone aware')
        return v


class Browser(BaseModel):
    name: str
    version: str
    comment: Optional[str]


class Creator(BaseModel):
    name: str
    version: str
    comment: Optional[str]


class HarLog(BaseModel):
    version: str
    creator: Creator
    browser: Optional[Browser]
    entries: List[Entry]
    comment: Optional[str]


class Har(BaseModel):
    log: HarLog


def _default_har_json_serialization(x):
    """
    Provides JSON serialization for HAR data structures that lack a default
    serialization.  For example, datetime elements need to be serialized as
    ISO 8601 strings.
    """
    if isinstance(x, datetime.datetime):
        return x.isoformat()
    else:
        return str(x)


class HarWriter:
    """
    Coordinates thread-safe request/response logging, encoded as an HTTP
    Archive (HAR) file.

    See http://www.softwareishard.com/har/viewer/ for the HAR formatting
    schema.
    """

    def __init__(self, *args):
        self.file = open(*args)
        self.finished = False
        self.first_entry = True
        self.queue = Queue()

        # Write initial data to the HAR file.
        self._write_preamble()

        Thread(name=f"HarWriter_{time.time}", target=self._writer).start()

    def write_entry(self, entry: Entry):
        sep = '' if self.first_entry else ', '
        self.first_entry = False

        self.queue.put(sep + json.dumps(entry.dict(exclude_defaults=True), default=_default_har_json_serialization))

    def close(self):
        # Close the entries array and the enclosing HAR JSON object.
        self.queue.join()
        self.finished = True
        self._write_postscript()
        self.file.close()

    def _write_preamble(self):
        creator = {
            "name": "Akita DjangoClient",
            "version": __version__,
            "comment": "See https://docs.akita.software.",
        }
        browser = {
            "name": "",
            "version": "",
        }
        comment = "Created by the Akita DjangoClient."
        self.file.write(f'''{{
            "log": {{
                "version": "1.2",
                "creator": {json.dumps(creator)},
                "browser": {json.dumps(browser)},
                "comment": "{comment}",
                "entries": [
        ''')

    def _write_postscript(self):
        # Terminate the entries array, log object, and enclosing HAR object.
        self.file.write(']}}')

    def _writer(self):
        while not self.finished:
            try:
                entry = self.queue.get(True, 1)
            except Empty:
                continue
            self.file.write(entry)
            self.queue.task_done()
